#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2023 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

import os

from migen import *
from migen.genlib.cdc import MultiReg

from litex.gen import *

from litex.soc.interconnect.csr import *
from litex.soc.cores.clock import S7MMCM

from litepcie.common import *
from litepcie.phy.common import *

# S7PCIEPHY ----------------------------------------------------------------------------------------

class S7PCIEPHY(LiteXModule):
    endianness    = "big"
    qword_aligned = False
    def __init__(self, platform, pads, data_width=64, bar0_size=1*MB, cd="sys", pcie_data_width=None):
        # Streams ----------------------------------------------------------------------------------
        self.sink   = stream.Endpoint(phy_layout(data_width))
        self.source = stream.Endpoint(phy_layout(data_width))
        self.msi    = stream.Endpoint(msi_layout())

        # Registers --------------------------------------------------------------------------------
        self._link_status = CSRStatus(fields=[
            CSRField("status", size=1, values=[
                ("``0b0``", "Link Down."),
                ("``0b1``", "Link Up."),
            ]),
            CSRField("rate", size=1, values=[
                ("``0b0``", "2.5 Gb/s."),
                ("``0b1``", "5.0 Gb/s."),
            ]),
            CSRField("width", size=2, values=[
                ("``0b00``", "1-Lane link."),
                ("``0b01``", "2-Lane link."),
                ("``0b10``", "4-Lane link."),
                ("``0b11``", "8-Lane link."),
            ]),
            CSRField("ltssm", size=6, description="LTSSM State"),
        ])
        self._msi_enable        = CSRStatus(description="MSI Enable Status. ``1``: MSI is enabled.")
        self._msix_enable       = CSRStatus(description="MSI-X Enable Status. ``1``: MSI-X is enabled.")
        self._bus_master_enable = CSRStatus(description="Bus Mastering Status. ``1``: Bus Mastering enabled.")
        self._max_request_size  = CSRStatus(16, description="Negiotiated Max Request Size (in bytes).")
        self._max_payload_size  = CSRStatus(16, description="Negiotiated Max Payload Size (in bytes).")

        # Parameters/Locals ------------------------------------------------------------------------
        if pcie_data_width is None: pcie_data_width = data_width
        self.platform         = platform
        self.data_width       = data_width
        self.pcie_data_width  = pcie_data_width

        self.id               = Signal(16, reset_less=True)
        self.bar0_size        = bar0_size
        self.bar0_mask        = get_bar_mask(bar0_size)
        self.max_request_size = Signal(16, reset_less=True)
        self.max_payload_size = Signal(16, reset_less=True)

        self.external_hard_ip = False

        # # #

        self.nlanes = nlanes = len(pads.tx_p)

        assert nlanes          in [1, 2, 4, 8]
        assert data_width      in [64, 128]
        assert pcie_data_width in [64, 128]

        # Clocking / Reset -------------------------------------------------------------------------
        self.pcie_refclk = pcie_refclk = Signal()
        self.pcie_rst_n  = pcie_rst_n  = Signal(reset=1)
        if hasattr(pads, "rst_n"):
            self.comb += pcie_rst_n.eq(pads.rst_n)
        self.specials += Instance("IBUFDS_GTE2",
            i_CEB = 0,
            i_I   = pads.clk_p,
            i_IB  = pads.clk_n,
            o_O   = pcie_refclk
        )
        platform.add_period_constraint(pads.clk_p, 1e9/100e6)
        self.cd_pcie = ClockDomain()

        # TX (FPGA --> HOST) CDC / Data Width Conversion -------------------------------------------
        self.tx_datapath = PHYTXDatapath(
            core_data_width = data_width,
            pcie_data_width = pcie_data_width,
            clock_domain    = cd)
        self.comb += self.sink.connect(self.tx_datapath.sink)
        s_axis_tx = self.tx_datapath.source

        # RX (HOST --> FPGA) CDC / Data Width Conversion -------------------------------------------
        self.rx_datapath = PHYRXDatapath(
            core_data_width = data_width,
            pcie_data_width = pcie_data_width,
            clock_domain    = cd,
            with_aligner    = True)
        m_axis_rx = self.rx_datapath.sink
        self.comb += self.rx_datapath.source.connect(self.source)

        # MSI CDC (FPGA --> HOST) ------------------------------------------------------------------
        if cd == "pcie":
            cfg_msi = self.msi
        else:
            self.msi_cdc = msi_cdc = stream.ClockDomainCrossing(
                layout          = msi_layout(),
                cd_from         = cd,
                cd_to           = "pcie",
                with_common_rst = True
            )
            self.comb += self.msi.connect(msi_cdc.sink)
            cfg_msi = msi_cdc.source

        # Hard IP Configuration --------------------------------------------------------------------

        def convert_size(command, size, max_size):
            cases = {}
            value = 128
            for i in range(6):
                cases[i] = size.eq(value)
                value = min(value*2, max_size)
            return Case(command, cases)

        bus_number      = Signal(8)
        device_number   = Signal(5)
        function_number = Signal(3)
        command         = Signal(16)
        dcommand        = Signal(16)
        self.sync.pcie += [
            convert_size(dcommand[12:15], self.max_request_size, max_size=512),
            convert_size(dcommand[5:8],   self.max_payload_size, max_size=512),
            self.id.eq(Cat(function_number, device_number, bus_number))
        ]
        self.specials += [
            MultiReg(command[2],            self._bus_master_enable.status),
            MultiReg(self.max_request_size, self._max_request_size.status),
            MultiReg(self.max_payload_size, self._max_payload_size.status)
        ]

        # Hard IP Clocking -------------------------------------------------------------------------

        # Signals.
        pipe_txoutclk      = Signal()
        pipe_txoutclk_bufg = Signal()
        pipe_pclk_sel      = Signal(nlanes)

        # Clock Domains.
        self.cd_clk125   = ClockDomain()
        self.cd_clk250   = ClockDomain()
        self.cd_userclk1 = ClockDomain()
        self.cd_userclk2 = ClockDomain()
        self.cd_pclk     = ClockDomain()

        # MMCM.
        userclk1_freq = {1:125e6, 2:125e6, 4:250e6, 8:500e6}[nlanes]
        userclk2_freq = {1:125e6, 2:125e6, 4:125e6, 8:250e6}[nlanes]
        self.mmcm = mmcm = S7MMCM(speedgrade=-2)
        self.specials += Instance("BUFG",
            i_I = pipe_txoutclk,
            o_O = pipe_txoutclk_bufg,
        )
        mmcm.register_clkin(pipe_txoutclk_bufg, 100e6)
        mmcm.create_clkout(self.cd_clk125,           125e6, margin=0)
        mmcm.create_clkout(self.cd_clk250,           250e6, margin=0)
        mmcm.create_clkout(self.cd_userclk1, userclk1_freq, margin=0)
        mmcm.create_clkout(self.cd_userclk2, userclk2_freq, margin=0)

        # PClk Selection.
        pipe_pclk_sel_r = Signal(nlanes)
        pclk_sel        = Signal()
        self.specials += MultiReg(pipe_pclk_sel, pipe_pclk_sel_r, "pclk")
        self.sync.pclk += [
            If(pipe_pclk_sel_r == (2**nlanes - 1), pclk_sel.eq(1)),
            If(pipe_pclk_sel_r ==               0, pclk_sel.eq(0)),
        ]
        self.specials += [
            Instance("BUFGCTRL",
                i_CE0 = 0b1,
                i_CE1 = 0b1,
                i_I0  = ClockSignal("clk125"),
                i_I1  = ClockSignal("clk250"),
                i_S0  = (pclk_sel == 0),
                i_S1  = (pclk_sel == 1),
                o_O   = ClockSignal("pclk"),
            )
        ]

        # Hard IP ----------------------------------------------------------------------------------
        m_axis_rx_tlast = Signal()
        m_axis_rx_tuser = Signal(32)

        self.pcie_phy_params = dict(
            # PCI Express Interface ----------------------------------------------------------------
            # Clk/Rst
            i_sys_clk     = pcie_refclk,
            i_sys_rst_n   = pcie_rst_n,

            # TX
            o_pci_exp_txp = pads.tx_p,
            o_pci_exp_txn = pads.tx_n,

            # RX
            i_pci_exp_rxp = pads.rx_p,
            i_pci_exp_rxn = pads.rx_n,

            # PIPE Clocking Interface --------------------------------------------------------------
            i_pipe_pclk_in      = ClockSignal("pclk"),
            o_pipe_txoutclk_out = pipe_txoutclk,
            o_pipe_rxoutclk_out = Open(),
            o_pipe_pclk_sel_out = pipe_pclk_sel,
            o_pipe_gen3_out     = Open(),
            i_pipe_rxusrclk_in  = ClockSignal("pclk"),
            i_pipe_rxoutclk_in  = 0,
            i_pipe_dclk_in      = ClockSignal("clk125"),
            i_pipe_userclk1_in  = ClockSignal("userclk1"),
            i_pipe_userclk2_in  = ClockSignal("userclk2"),
            i_pipe_oobclk_in    = ClockSignal("pclk"),
            i_pipe_mmcm_lock_in = mmcm.locked,
            i_pipe_mmcm_rst_n   = 1,

            # AXI-S Interface ----------------------------------------------------------------------
            # Common
            o_user_clk_out     = ClockSignal("pcie"),
            o_user_reset_out   = ResetSignal("pcie"),
            o_user_lnk_up      = self._link_status.fields.status,
            o_user_app_rdy     = Open(),

            # TX
            o_tx_buf_av        = Open(),
            o_tx_err_drop      = Open(),
            o_tx_cfg_req       = Open(),
            i_tx_cfg_gnt       = 1,
            i_s_axis_tx_tvalid = s_axis_tx.valid,
            i_s_axis_tx_tlast  = s_axis_tx.last,
            o_s_axis_tx_tready = s_axis_tx.ready,
            i_s_axis_tx_tdata  = s_axis_tx.dat,
            i_s_axis_tx_tkeep  = s_axis_tx.be,
            i_s_axis_tx_tuser  = 0,

            # RX
            i_rx_np_ok         = 1,
            i_rx_np_req        = 1,
            o_m_axis_rx_tvalid = m_axis_rx.valid,
            o_m_axis_rx_tlast  = m_axis_rx_tlast,
            i_m_axis_rx_tready = m_axis_rx.ready,
            o_m_axis_rx_tdata  = m_axis_rx.dat,
            o_m_axis_rx_tkeep  = m_axis_rx.be,
            o_m_axis_rx_tuser  = m_axis_rx_tuser,

            # Flow Control
            o_fc_cpld          = Open(),
            o_fc_cplh          = Open(),
            o_fc_npd           = Open(),
            o_fc_nph           = Open(),
            o_fc_pd            = Open(),
            o_fc_ph            = Open(),
            i_fc_sel           = 0,

            # Management Interface -----------------------------------------------------------------
            o_cfg_mgmt_do            = Open(),
            o_cfg_mgmt_rd_wr_done    = Open(),
            i_cfg_mgmt_di            = 0,
            i_cfg_mgmt_byte_en       = 0,
            i_cfg_mgmt_dwaddr        = 0,
            i_cfg_mgmt_wr_en         = 0,
            i_cfg_mgmt_rd_en         = 0,
            i_cfg_mgmt_wr_readonly   = 0,
            i_cfg_mgmt_wr_rw1c_as_rw = 0,

            # Error Reporting Interface ------------------------------------------------------------
            i_cfg_err_ecrc                  = 0,
            i_cfg_err_ur                    = 0,
            i_cfg_err_cpl_timeout           = 0,
            i_cfg_err_cpl_unexpect          = 0,
            i_cfg_err_cpl_abort             = 0,
            i_cfg_err_posted                = 0,
            i_cfg_err_cor                   = 0,
            i_cfg_err_atomic_egress_blocked = 0,
            i_cfg_err_internal_cor          = 0,
            i_cfg_err_malformed             = 0,
            i_cfg_err_mc_blocked            = 0,
            i_cfg_err_poisoned              = 0,
            i_cfg_err_norecovery            = 0,
            i_cfg_err_tlp_cpl_header        = 0,
            o_cfg_err_cpl_rdy               = Open(),
            i_cfg_err_locked                = 0,
            i_cfg_err_acs                   = 0,
            i_cfg_err_internal_uncor        = 0,

            # AER interface ------------------------------------------------------------------------
            i_cfg_err_aer_headerlog     = 0,
            i_cfg_aer_interrupt_msgnum  = 0,
            o_cfg_err_aer_headerlog_set = Open(),
            o_cfg_aer_ecrc_check_en     = Open(),
            o_cfg_aer_ecrc_gen_en       = Open(),

            i_cfg_turnoff_ok            = 0,
            i_cfg_trn_pending           = 0,
            i_cfg_pm_halt_aspm_l0s      = 0,
            i_cfg_pm_halt_aspm_l1       = 0,
            i_cfg_pm_force_state_en     = 0,
            i_cfg_pm_force_state        = 0,
            i_cfg_dsn                   = 0,
            i_cfg_pm_send_pme_to        = 0,
            i_cfg_ds_bus_number         = 0,
            i_cfg_ds_device_number      = 0,
            i_cfg_ds_function_number    = 0,
            i_cfg_pm_wake               = 0,

            # Interrupt Interface ------------------------------------------------------------------
            i_cfg_interrupt                = cfg_msi.valid,
            o_cfg_interrupt_rdy            = cfg_msi.ready,
            i_cfg_interrupt_assert         = 0,
            i_cfg_interrupt_di             = cfg_msi.dat,
            o_cfg_interrupt_do             = Open(),
            o_cfg_interrupt_mmenable       = Open(),
            o_cfg_interrupt_msienable      = self._msi_enable.status,
            o_cfg_interrupt_msixenable     = self._msix_enable.status,
            o_cfg_interrupt_msixfm         = Open(),
            i_cfg_interrupt_stat           = 0,
            i_cfg_pciecap_interrupt_msgnum = 0,

            # Configuration Interface --------------------------------------------------------------
            o_cfg_status                                 = Open(),
            o_cfg_command                                = command,
            o_cfg_dstatus                                = Open(),
            o_cfg_dcommand                               = dcommand,
            o_cfg_lstatus                                = Open(),
            o_cfg_lcommand                               = Open(),
            o_cfg_dcommand2                              = Open(),
            o_cfg_pcie_link_state                        = Open(),
            o_cfg_to_turnoff                             = Open(),
            o_cfg_bus_number                             = bus_number,
            o_cfg_device_number                          = device_number,
            o_cfg_function_number                        = function_number,

            o_cfg_pmcsr_pme_en                           = Open(),
            o_cfg_pmcsr_powerstate                       = Open(),
            o_cfg_pmcsr_pme_status                       = Open(),
            o_cfg_received_func_lvl_rst                  = Open(),
            o_cfg_bridge_serr_en                         = Open(),
            o_cfg_slot_control_electromech_il_ctl_pulse  = Open(),
            o_cfg_root_control_syserr_corr_err_en        = Open(),
            o_cfg_root_control_syserr_non_fatal_err_en   = Open(),
            o_cfg_root_control_syserr_fatal_err_en       = Open(),
            o_cfg_root_control_pme_int_en                = Open(),
            o_cfg_aer_rooterr_corr_err_reporting_en      = Open(),
            o_cfg_aer_rooterr_non_fatal_err_reporting_en = Open(),
            o_cfg_aer_rooterr_fatal_err_reporting_en     = Open(),
            o_cfg_aer_rooterr_corr_err_received          = Open(),
            o_cfg_aer_rooterr_non_fatal_err_received     = Open(),
            o_cfg_aer_rooterr_fatal_err_received         = Open(),

            # VC Interface -------------------------------------------------------------------------
            o_cfg_vc_tcvc_map                    = Open(),

            o_cfg_msg_received                   = Open(),
            o_cfg_msg_data                       = Open(),
            o_cfg_msg_received_pm_as_nak         = Open(),
            o_cfg_msg_received_setslotpowerlimit = Open(),
            o_cfg_msg_received_err_cor           = Open(),
            o_cfg_msg_received_err_non_fatal     = Open(),
            o_cfg_msg_received_err_fatal         = Open(),
            o_cfg_msg_received_pm_pme            = Open(),
            o_cfg_msg_received_pme_to_ack        = Open(),
            o_cfg_msg_received_assert_int_a      = Open(),
            o_cfg_msg_received_assert_int_b      = Open(),
            o_cfg_msg_received_assert_int_c      = Open(),
            o_cfg_msg_received_assert_int_d      = Open(),
            o_cfg_msg_received_deassert_int_a    = Open(),
            o_cfg_msg_received_deassert_int_b    = Open(),
            o_cfg_msg_received_deassert_int_c    = Open(),
            o_cfg_msg_received_deassert_int_d    = Open(),


            # Physical Layer Interface -------------------------------------------------------------
            i_pl_directed_link_change        = 0,
            i_pl_directed_link_width         = 0,
            i_pl_directed_link_speed         = 0,
            i_pl_directed_link_auton         = 0,
            i_pl_upstream_prefer_deemph      = 1,
            o_pl_sel_lnk_rate                = self._link_status.fields.rate,
            o_pl_sel_lnk_width               = self._link_status.fields.width,
            o_pl_ltssm_state                 = self._link_status.fields.ltssm,
            o_pl_lane_reversal_mode          = Open(),
            o_pl_phy_lnk_up                  = Open(),
            o_pl_tx_pm_state                 = Open(),
            o_pl_rx_pm_state                 = Open(),
            o_pl_link_upcfg_cap              = Open(),
            o_pl_link_gen2_cap               = Open(),
            o_pl_link_partner_gen2_supported = Open(),
            o_pl_initial_link_width          = Open(),
            o_pl_directed_change_done        = Open(),
            o_pl_received_hot_rst            = Open(),
            i_pl_transmit_hot_rst            = 0,
            i_pl_downstream_deemph_source    = 0,

            # PCIe DRP Interface -------------------------------------------------------------------
            i_pcie_drp_clk  = 1,
            i_pcie_drp_en   = 0,
            i_pcie_drp_we   = 0,
            i_pcie_drp_addr = 0,
            i_pcie_drp_di   = 0,
            o_pcie_drp_rdy  = Open(),
            o_pcie_drp_do   = Open(),
        )
        if pcie_data_width == 128:
            rx_is_sof = m_axis_rx_tuser[10:15] # Start of a new packet header in m_axis_rx_tdata.
            rx_is_eof = m_axis_rx_tuser[17:22] # End of a packet in m_axis_rx_tdata.
            self.comb += [
                m_axis_rx.first.eq(rx_is_sof[-1]),
                m_axis_rx.last.eq( rx_is_eof[-1]),
                If(rx_is_sof == 0b11000, self.rx_datapath.aligner.first_dword.eq(2)),
            ]
        else:
            self.comb += [
                m_axis_rx.first.eq(0),
                m_axis_rx.last.eq(m_axis_rx_tlast),
            ]

    # LTSSM Tracer ---------------------------------------------------------------------------------
    def add_ltssm_tracer(self):
        self.ltssm_tracer = LTSSMTracer(self._link_status.fields.ltssm)

    # Hard IP sources ------------------------------------------------------------------------------
    def add_sources(self, platform, phy_path, phy_filename=None):
        if phy_filename is not None:
            platform.add_ip(os.path.join(phy_path, phy_filename))
        else:
            config = {
                "Bar0_Scale"         : "Megabytes",
                "Bar0_Size"          : 1,
                "Buf_Opt_BMA"        : True,
                "Component_Name"     : "pcie",
                "Device_ID"          : 7020 + self.nlanes,
                "IntX_Generation"    : False,
                "Interface_Width"    : f"{self.pcie_data_width}_bit",
                "Legacy_Interrupt"   : None,
                "Multiple_Message_Capable"  : '1_vector',
                "Link_Speed"         : "5.0_GT/s",
                "MSI_64b"            : False,
                "Max_Payload_Size"   : "512_bytes" if self.nlanes != 8 else "256_bytes",
                "Maximum_Link_Width" : f"X{self.nlanes}",
                "PCIe_Blk_Locn"      : "X0Y0",
                "Ref_Clk_Freq"       : "100_MHz",
                "Trans_Buf_Pipeline" : None,
                "Trgt_Link_Speed"    : "4'h2",
                "User_Clk_Freq"      : 125 if self.nlanes != 8 else 250,
            }
            ip_tcl = []
            ip_tcl.append("create_ip -vendor xilinx.com -name pcie_7x -module_name pcie_s7")
            ip_tcl.append("set obj [get_ips pcie_s7]")
            ip_tcl.append("set_property -dict [list \\")
            for config, value in config.items():
                ip_tcl.append("CONFIG.{} {} \\".format(config, '{{' + str(value) + '}}'))
            ip_tcl.append(f"] $obj")
            ip_tcl.append("synth_ip $obj")
            platform.toolchain.pre_synthesis_commands += ip_tcl
        # Reset LOC constraints on GTPE2_COMMON and BRAM36 from .xci (we only want to keep Timing constraints).
        if platform.device.startswith("xc7a"):
            platform.toolchain.pre_placement_commands.append("reset_property LOC [get_cells -hierarchical -filter {{NAME=~pcie_s7/*gtp_common.gtpe2_common_i}}]")
        else:
            platform.toolchain.pre_placement_commands.append("reset_property LOC [get_cells -hierarchical -filter {{NAME=~pcie_s7/*gtx_common.gtxe2_common_i}}]")
        if self.nlanes != 8:
            platform.toolchain.pre_placement_commands.append("reset_property LOC [get_cells -hierarchical -filter {{NAME=~pcie_s7/*genblk*.bram36_tdp_bl.bram36_tdp_bl}}]")

    # External Hard IP -----------------------------------------------------------------------------
    def use_external_hard_ip(self, hard_ip_path, hard_ip_filename):
        self.external_hard_ip = True
        self.add_sources(self.platform, hard_ip_path, hard_ip_filename)

    # Finalize -------------------------------------------------------------------------------------
    def do_finalize(self):
        if not self.external_hard_ip:
            phy_path     = "xilinx_s7_gen2"
            self.add_sources(self.platform,
                phy_path     = os.path.join(os.path.abspath(os.path.dirname(__file__)), phy_path),
            )
        self.specials += Instance("pcie_s7", **self.pcie_phy_params)



#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2022 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.soc.interconnect.csr import *
from litex.soc.interconnect.csr_bus import SRAM

from litepcie.common import *

# LitePCIeMSI --------------------------------------------------------------------------------------

class LitePCIeMSI(Module, AutoCSR):
    def __init__(self, width=32):
        self.irqs   = Signal(width)
        self.source = stream.Endpoint(msi_layout())

        self.enable = CSRStorage(width, description="""MSI Enable Control.\n
           Write bit(s) to ``1`` to enable corresponding MSI IRQ(s).""")
        self.clear  = CSRStorage(width, description="""MSI Clear Control.\n
           Write bit(s) to ``1`` to clear corresponding MSI IRQ(s).""")
        self.vector = CSRStatus(width,  description="""MSI Vector Status.\n
           Current MSI IRQs vector value.""")

        # # #

        enable = Signal(width)
        clear  = Signal(width)
        vector = Signal(width)

        # Memorize and clear IRQ Vector ------------------------------------------------------------
        self.comb += If(self.clear.re, clear.eq(self.clear.storage))
        self.comb += enable.eq(self.enable.storage)
        self.comb += self.vector.status.eq(vector)
        self.sync += vector.eq(enable & ((vector & ~clear) | self.irqs))

        # Generate MSI -----------------------------------------------------------------------------
        msi = Signal(width)
        self.comb += self.source.valid.eq(msi != 0)
        self.sync += [
            msi.eq((msi | self.irqs) & enable),
            If(self.source.ready,
                msi.eq(self.irqs & enable)
            )
        ]

# LitePCIeMSIMultiVector ---------------------------------------------------------------------------

class LitePCIeMSIMultiVector(Module, AutoCSR):
  def __init__(self, width=32):
        self.irqs   = Signal(width)
        self.source = stream.Endpoint(msi_layout())

        self.enable = CSRStorage(width, description="""MSI Enable Control.\n
           Write bit(s) to ``1`` to enable corresponding MSI IRQ(s).""")

        # # #

        enable = Signal(width)
        clear  = Signal(width)
        vector = Signal(width)

        # Memorize and clear IRQ Vector ------------------------------------------------------------
        self.comb += enable.eq(self.enable.storage)
        self.sync += vector.eq(enable & ((vector & ~clear) | self.irqs))

        # Generate MSI -----------------------------------------------------------------------------
        for i in reversed(range(width)): # Priority given to lower indexes.
            self.comb += [
                If(vector[i],
                    self.source.valid.eq(1),
                    self.source.dat.eq(i),
                    If(self.source.ready,
                        clear.eq(1 << i)
                    )
                )
            ]

# LitePCIeMSIX -------------------------------------------------------------------------------------

class LitePCIeMSIX(Module, AutoCSR):
    def __init__(self, endpoint, width=32):
        assert width <= 64
        self.irqs           = Signal(width)
        self.enable         = CSRStorage(width, description="""MSI-X Enable Control.\n
           Write bit(s) to ``1`` to enable corresponding MSI-X IRQ(s).""")
        if width <= 32:
            self.reserved0 = CSRStorage() # For 64-bit alignment.
        self.pba            = CSRStatus(width, description="""MSI-X PBA Table.""")
        if width <= 32:
            self.reserved1 = CSRStorage() # For 64-bit alignment.
        self.specials.table = Memory(4*32, width) # MSI-X Table.

        # # #

        enable = Signal(width)
        clear  = Signal(width)
        vector = Signal(width)

        # Memorize and clear IRQ Vector ------------------------------------------------------------
        self.comb += enable.eq(self.enable.storage)
        self.sync += vector.eq(enable & ((vector & ~clear) | self.irqs))
        self.comb += self.pba.status.eq(vector)

        # Generate MSI-X ---------------------------------------------------------------------------
        msix_valid = Signal()
        msix_ready = Signal()
        msix_num   = Signal(max=width)

        for i in reversed(range(width)): # Priority given to lower indexes.
            self.comb += [
                If(vector[i],
                    msix_valid.eq(1),
                    msix_num.eq(i),
                    If(msix_ready,
                        clear.eq(1 << i)
                    )
                )
            ]

        # Send MSI-X as TLP-Write ------------------------------------------------------------------
        port     = endpoint.crossbar.get_master_port()
        table_port = self.table.get_port(has_re=True)
        self.specials += table_port

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            table_port.adr.eq(msix_num),
            table_port.re.eq(1),
            If(msix_valid,
                NextState("ISSUE-WRITE")
            )
        )
        self.comb += [
            port.source.channel.eq(port.channel),
            port.source.first.eq(1),
            port.source.last.eq(1),
            port.source.adr.eq(table_port.dat_r[96:128]), # Lower Address from table.
            port.source.req_id.eq(endpoint.phy.id),
            port.source.tag.eq(0),
            port.source.len.eq(1),
            port.source.dat.eq(table_port.dat_r[32:64]), # Message Data from table.
        ]
        fsm.act("ISSUE-WRITE",
            port.source.valid.eq(1),
            port.source.we.eq(1),
            If(port.source.ready,
                msix_ready.eq(1),
                NextState("IDLE")
            )
        )



#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2020 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *
from migen.genlib.cdc import MultiReg

from litepcie.common import *

# TX Datapath --------------------------------------------------------------------------------------

class PHYTXDatapath(Module):
    def __init__(self, core_data_width, pcie_data_width, clock_domain):
        self.sink   = sink   = stream.Endpoint(phy_layout(core_data_width))
        self.source = source = stream.Endpoint(phy_layout(pcie_data_width))

        # # #

        if (clock_domain == "pcie") and (core_data_width == pcie_data_width):
            self.comb += sink.connect(source)
        else:
            pipe_valid = stream.PipeValid(phy_layout(core_data_width))
            pipe_valid = ClockDomainsRenamer(clock_domain)(pipe_valid)
            cdc        = stream.ClockDomainCrossing(
                layout          = phy_layout(core_data_width),
                cd_from         = clock_domain,
                cd_to           = "pcie",
                with_common_rst = True,
                depth           = 16,
            )
            converter  = stream.StrideConverter(phy_layout(core_data_width), phy_layout(pcie_data_width))
            converter  = ClockDomainsRenamer("pcie")(converter)
            pipe_ready = stream.PipeReady(phy_layout(pcie_data_width))
            pipe_ready = ClockDomainsRenamer("pcie")(pipe_ready)
            self.submodules += pipe_valid, cdc, converter, pipe_ready
            self.comb += [
                sink.connect(pipe_valid.sink),
                pipe_valid.source.connect(cdc.sink),
                cdc.source.connect(converter.sink),
                converter.source.connect(pipe_ready.sink),
                pipe_ready.source.connect(source),
            ]

# PHYRX128BAligner ---------------------------------------------------------------------------------

class PHYRX128BAligner(Module):
    def __init__(self):
        self.sink   = sink   = stream.Endpoint(phy_layout(128))
        self.source = source = stream.Endpoint(phy_layout(128))
        self.first_dword = Signal(2)

        # # #

        dat_last = Signal(64, reset_less=True)
        be_last  = Signal(8,  reset_less=True)
        self.sync += [
            If(sink.valid & sink.ready,
                dat_last.eq(sink.dat[64:]),
                be_last.eq( sink.be[8:]),
            )
        ]

        self.submodules.fsm = fsm = FSM(reset_state="ALIGNED")
        fsm.act("ALIGNED",
            sink.connect(source, omit={"first"}),
            # If "first" on DWORD2 and "last" on the same cycle, switch to UNALIGNED.
            If(sink.valid & sink.last & sink.first & (self.first_dword == 2),
                source.be[8:].eq(0),
                If(source.ready,
                    NextState("UNALIGNED")
                )
            )
        )
        fsm.act("UNALIGNED",
            sink.connect(source, omit={"first", "dat", "be"}),
            source.dat.eq(Cat(dat_last, sink.dat)),
            source.be.eq( Cat(be_last,  sink.be)),
            # If "last" and not "first" on the same cycle, switch to ALIGNED.
            If(sink.valid & sink.last & ~sink.first,
                source.be[8:].eq(0),
                If(source.ready,
                    NextState("ALIGNED")
                )
            )
        )

# RX Datapath --------------------------------------------------------------------------------------

class PHYRXDatapath(Module):
    def __init__(self, core_data_width, pcie_data_width, clock_domain, with_aligner=False):
        self.sink   = sink   = stream.Endpoint(phy_layout(pcie_data_width))
        self.source = source = stream.Endpoint(phy_layout(core_data_width))

        # # #

        if (pcie_data_width == 128) and with_aligner:
            aligner = PHYRX128BAligner()
            aligner = ClockDomainsRenamer("pcie")(aligner)
            self.submodules.aligner = aligner
            self.comb += sink.connect(aligner.sink)
            sink = aligner.source

        if (clock_domain == "pcie") and (core_data_width == pcie_data_width):
            self.comb += sink.connect(source)
        else:
            pipe_ready = stream.PipeReady(phy_layout(core_data_width))
            pipe_ready = ClockDomainsRenamer("pcie")(pipe_ready)
            converter  = stream.StrideConverter(phy_layout(pcie_data_width), phy_layout(core_data_width))
            converter  = ClockDomainsRenamer("pcie")(converter)
            cdc        = stream.ClockDomainCrossing(
                layout          = phy_layout(core_data_width),
                cd_from         = "pcie",
                cd_to           = clock_domain,
                with_common_rst = True,
                depth           = 16,
            )
            pipe_valid = stream.PipeValid(phy_layout(core_data_width))
            pipe_valid = ClockDomainsRenamer(clock_domain)(pipe_valid)
            self.submodules += pipe_ready, converter, cdc, pipe_valid
            self.comb += [
                sink.connect(pipe_ready.sink),
                pipe_ready.source.connect(converter.sink),
                converter.source.connect(cdc.sink),
                cdc.source.connect(pipe_valid.sink),
                pipe_valid.source.connect(source),
            ]

# LTSSMTracer --------------------------------------------------------------------------------------

class LTSSMTracer(Module, AutoCSR):
    def __init__(self, ltssm):
        self._history = CSRStatus(description="History of LTSSM states",
            fields = [
                CSRField("new",   offset= 0, size=6, description="New LTSSM state"),
                CSRField("old",   offset= 6, size=6, description="Old LTSSM state"),
                CSRField("ovfl",  offset=30, size=1, description="Overflow"),
                CSRField("valid", offset=31, size=1, description="Is data valid"),
        ])

        # The ltssm state signal input is sampled in the sys domain just using a MultiReg. This means
        # on change we could have an invalid state during 1 cycle.
        #
        # We also don't use an AsyncFIFO because the pcie clock domain is held in reset most of the
        # LTSSM initial negotiation defeating the point of this module.

        fifo = stream.SyncFIFO([("new", 6), ("old", 6), ("ovfl", 1)], 128)
        self.submodules += fifo

        ltssm_cur = Signal(6)
        ltssm_d1  = Signal(6)
        ltssm_d2  = Signal(6)

        overflow  = Signal()
        change    = Signal()

        self.sync += [
            ltssm_d1.eq(ltssm_cur),
            If(fifo.sink.ready, ltssm_d2.eq(ltssm_d1)),
            change.eq((ltssm_cur != ltssm_d1) & ~change),
            overflow.eq((overflow | change) & ~fifo.sink.ready),
        ]

        self.comb += [
            ltssm_cur.eq(ltssm),
            fifo.sink.new.eq(ltssm_cur),
            fifo.sink.old.eq(ltssm_d2),
            fifo.sink.ovfl.eq(overflow),
            fifo.sink.valid.eq(overflow | change),
            self._history.fields.new.eq(fifo.source.new),
            self._history.fields.old.eq(fifo.source.old),
            self._history.fields.ovfl.eq(fifo.source.ovfl),
            self._history.fields.valid.eq(fifo.source.valid),
            fifo.source.ready.eq(self._history.we),
        ]



#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2022 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litepcie.common import *
from litepcie.core.common import *
from litepcie.tlp.controller import LitePCIeTLPController

# LitePCIe Crossbar --------------------------------------------------------------------------------

class LitePCIeCrossbar(Module):
    def __init__(self, data_width, address_width, max_pending_requests, cmp_bufs_buffered=True):
        self.data_width           = data_width
        self.address_width        = address_width
        self.max_pending_requests = max_pending_requests
        self.cmp_bufs_buffered    = cmp_bufs_buffered

        self.master     = LitePCIeMasterInternalPort(data_width, address_width)
        self.slave      = LitePCIeSlaveInternalPort(data_width)
        self.phy_master = LitePCIeMasterPort(self.master)
        self.phy_slave  = LitePCIeSlavePort(self.slave)

        self.user_masters         = []
        self.user_masters_channel = 0
        self.user_slaves          = []

    def get_slave_port(self, address_decoder):
        s = LitePCIeSlaveInternalPort(
            data_width      = self.data_width,
            address_decoder = address_decoder)
        self.user_slaves.append(s)
        return LitePCIeSlavePort(s)

    def get_master_port(self, write_only=False, read_only=False):
        m = LitePCIeMasterInternalPort(
            data_width    = self.data_width,
            address_width = self.address_width,
            channel       = self.user_masters_channel,
            write_only    = write_only,
            read_only     = read_only
        )
        self.user_masters_channel += 1
        self.user_masters.append(m)
        return LitePCIeMasterPort(m)

    def filter_masters(self, write_only, read_only):
        masters = []
        for m in self.user_masters:
            if m.write_only == write_only and m.read_only == read_only:
                masters.append(m)
        return masters

    def slave_dispatch_arbitrate(self, slaves, slave):
        # Dispatch ---------------------------------------------------------------------------------
        s_sources    = [s.source for s in slaves]
        s_dispatcher = Dispatcher(slave.source, s_sources, one_hot=True)
        self.submodules += s_dispatcher
        for i, s in enumerate(slaves):
                self.comb += s_dispatcher.sel[i].eq(s.address_decoder(slave.source.adr))

        # Arbitrate --------------------------------------------------------------------------------
        s_sinks   = [s.sink for s in slaves]
        s_arbiter = Arbiter(s_sinks, slave.sink)
        self.submodules += s_arbiter

    def master_arbitrate_dispatch(self, masters, master, dispatch=True):
        # Arbitrate --------------------------------------------------------------------------------
        m_sinks   = [m.sink for m in masters]
        m_arbiter = Arbiter(m_sinks, master.sink)
        self.submodules += m_arbiter

        # Dispatch ---------------------------------------------------------------------------------
        if dispatch:
            m_sources    = [m.source for m in masters]
            m_dispatcher = Dispatcher(master.source, m_sources, one_hot=True)
            self.submodules += m_dispatcher
            for i, m in enumerate(masters):
                if m.channel is not None:
                    self.comb += m_dispatcher.sel[i].eq(master.source.channel == m.channel)
        else:
            # Connect to first master
            self.comb += master.source.connect(masters[0].source)

    def do_finalize(self):
        # Slave path -------------------------------------------------------------------------------
        # Dispatch request to user sources (according to address decoder)
        # Arbitrate completion from user sinks
        if self.user_slaves != []:
            self.slave_dispatch_arbitrate(self.user_slaves, self.slave)

        # Master path ------------------------------------------------------------------------------
        # Abritrate requests from user sinks
        # Dispatch completion to user sources (according to channel)

        #           +-------+
        #  reqs---> |  RD   |
        #  cmps<--- | PORTS |---------+
        #           +-------+     +---+----+   +----------+
        #                         |Arb/Disp|-->|Controller|--+
        #           +-------+     +---+----+   +----------+  |
        #  reqs---> |  RW   |         |                      |
        #  cmps<--- | PORTS |---------+                      |
        #           +-------+                            +---+----+
        #                                                |Arb/Disp|<--> to/from  Packetizer/
        #           +-------+                            +---+----+              Depacketizer
        #  reqs---> |  WR   |     +--------+                 |
        #  cmps<--- | PORTS |-----|Arb/Disp|-----------------+
        #           +-------+     +--------+
        #
        # The controller blocks RD requests when the max number of pending
        # requests have been sent (max_pending_requests parameters).
        # To avoid blocking write_only ports when RD requests are blocked,
        # a separate arbitration stage is used.

        if self.user_masters != []:
            masters = []

            # Arbitrate / dispatch read_only / read_write ports ------------------------------------
            # and insert controller
            rd_rw_masters = self.filter_masters(False, True)
            rd_rw_masters += self.filter_masters(False, False)
            if rd_rw_masters != []:
                rd_rw_master = LitePCIeMasterInternalPort(self.data_width, self.address_width)
                controller   = LitePCIeTLPController(
                    data_width           = self.data_width,
                    address_width        = self.address_width,
                    max_pending_requests = self.max_pending_requests,
                    cmp_bufs_buffered    = self.cmp_bufs_buffered)
                self.submodules.controller = controller
                self.master_arbitrate_dispatch(rd_rw_masters, controller.master_in)
                masters.append(controller.master_out)

            # Arbitrate / dispatch write_only ports ------------------------------------------------
            wr_masters = self.filter_masters(True, False)
            if wr_masters != []:
                wr_master = LitePCIeMasterInternalPort(self.data_width, self.address_width)
                self.master_arbitrate_dispatch(wr_masters, wr_master)
                masters.append(wr_master)

            # Final Arbitrate / dispatch stage -----------------------------------------------------
            self.master_arbitrate_dispatch(masters, self.master, False)



#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2022 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litepcie.common import *

# LitePCIe Internal Ports --------------------------------------------------------------------------

class LitePCIeSlaveInternalPort:
    def __init__(self, data_width, address_width=32, address_decoder=None):
        self.address_decoder = address_decoder
        self.sink   = stream.Endpoint(completion_layout(data_width))
        self.source = stream.Endpoint(request_layout(data_width))


class LitePCIeMasterInternalPort:
    def __init__(self, data_width, address_width=32, channel=None, write_only=False, read_only=False):
        self.channel    = channel
        self.write_only = write_only
        self.read_only  = read_only
        self.sink   = stream.Endpoint(request_layout(data_width, address_width))
        self.source = stream.Endpoint(completion_layout(data_width))

# LitePCIe User Ports ------------------------------------------------------------------------------

class LitePCIeSlavePort:
    def __init__(self, port):
        self.address_decoder = port.address_decoder
        self.sink   = port.source
        self.source = port.sink


class LitePCIeMasterPort:
    def __init__(self, port):
        self.channel = port.channel
        self.sink    = port.source
        self.source  = port.sink



#
# This file is part of LitePCIe.
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.soc.interconnect import axi, stream

from litepcie.common import *

from litepcie.frontend.dma import descriptor_layout, LitePCIeDMAWriter, LitePCIeDMAReader

# LitePCIeAXISlave ---------------------------------------------------------------------------------

class LitePCIeAXISlave(Module):
    def __init__(self, endpoint, data_width=32, id_width=1):
        self.axi = axi.AXIInterface(data_width=data_width, id_width=id_width)

        # # #

        aw_id = Signal(id_width)
        ar_id = Signal(id_width)
        r_len = Signal(8)

        desc_rd = stream.Endpoint(descriptor_layout())
        desc_wr = stream.Endpoint(descriptor_layout())

        port_rd = endpoint.crossbar.get_master_port(read_only=True)
        port_wr = endpoint.crossbar.get_master_port(write_only=True)

        # AXI Write Path ---------------------------------------------------------------------------

        # DMA / FIFO / Converter
        self.submodules.dma_wr = dma_wr = LitePCIeDMAWriter(
            endpoint = endpoint,
            port     = port_wr,
            with_table = False)
        self.submodules.fifo_wr = fifo_wr = stream.SyncFIFO(descriptor_layout(), 16)
        self.submodules.conv_wr = conv_wr = stream.Converter(nbits_from=data_width, nbits_to=endpoint.phy.data_width)

        # Flow
        self.comb += [
            desc_wr.connect(fifo_wr.sink),
            fifo_wr.source.connect(dma_wr.desc_sink),
            conv_wr.source.connect(dma_wr.sink),
        ]

        # FSM (Convert AXI Write Requests to LitePCIe's DMA Descriptors).
        self.comb += desc_wr.address.eq(self.axi.aw.addr)                       # Start address (byte addressed)
        self.comb += desc_wr.length.eq((self.axi.aw.len + 1) * (data_width//8)) # Transfer length (in bytes)

        self.submodules.fsm_wr = fsm_wr = FSM(reset_state="WRITE-IDLE")
        fsm_wr.act("WRITE-IDLE",
            self.axi.aw.ready.eq(desc_wr.ready),
            desc_wr.valid.eq(self.axi.aw.valid),
            If(self.axi.aw.valid & self.axi.aw.ready,
                NextValue(aw_id, self.axi.aw.id), # Save id to use it on b channel.
                NextState("WRITE-MONITOR"),
            )
        )

        self.comb += [
            conv_wr.sink.data.eq(self.axi.w.data),
            conv_wr.sink.last.eq(self.axi.w.last),
        ]
        fsm_wr.act("WRITE-MONITOR",
            conv_wr.sink.valid.eq(self.axi.w.valid),
            self.axi.w.ready.eq(conv_wr.sink.ready),
            If(self.axi.w.valid & self.axi.w.ready & self.axi.w.last,
                NextState("WRITE-RESP"),
            )
        )

        self.comb += [
            self.axi.b.id.eq(aw_id),
            self.axi.b.resp.eq(0),
        ]
        fsm_wr.act("WRITE-RESP",
            self.axi.b.valid.eq(1),
            If(self.axi.b.ready,
                NextState("WRITE-IDLE"), # Write done
            )
        )

        # AXI Read Path ----------------------------------------------------------------------------

        # DMA / FIFO / Converter
        self.submodules.dma_rd = dma_rd = LitePCIeDMAReader(
            endpoint = endpoint,
            port     = port_rd,
            with_table = False)
        self.submodules.fifo_rd = fifo_rd = stream.SyncFIFO(descriptor_layout(), 16)
        self.submodules.conv_rd = conv_rd = stream.Converter(nbits_from=endpoint.phy.data_width, nbits_to=data_width)

        # Flow
        self.comb += [
            desc_rd.connect(fifo_rd.sink),
            fifo_rd.source.connect(dma_rd.desc_sink),
            dma_rd.source.connect(conv_rd.sink),
        ]

        # FSM (Convert AXI Read Requests to LitePCIe's DMA Descriptors).
        self.comb += desc_rd.address.eq(self.axi.ar.addr)                       # Starting address (byte addressed)
        self.comb += desc_rd.length.eq((self.axi.ar.len + 1) * (data_width//8)) # Transfer length (in bytes)

        self.submodules.fsm_rd = fsm_rd = FSM(reset_state="READ-IDLE")
        fsm_rd.act("READ-IDLE",
            self.axi.ar.ready.eq(desc_rd.ready),
            desc_rd.valid.eq(self.axi.ar.valid),
            If(self.axi.ar.valid & self.axi.ar.ready,
                NextValue(ar_id, self.axi.ar.id), # Save id to use it on r channel.
                NextValue(r_len, self.axi.ar.len),
                NextState("READ-MONITOR"),
            )
        )

        self.comb += [
            self.axi.r.data.eq(conv_rd.source.data),
            self.axi.r.last.eq(r_len == 0),
            # We need to provide the same id that was provided on aw channel for the duration of the transfer.
            self.axi.r.id.eq(ar_id),
            self.axi.r.resp.eq(0),
        ]
        fsm_rd.act("READ-MONITOR",
            self.axi.r.valid.eq(conv_rd.source.valid),
            conv_rd.source.ready.eq(self.axi.r.ready),
            If(self.axi.r.ready & self.axi.r.valid,
                NextValue(r_len, r_len - 1),
                If(self.axi.r.last, # Check if we finished the whole AXI transaction.
                    NextState("READ-IDLE"),
                )
            )
        )




#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2022 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litepcie.common import *

# Constants ----------------------------------------------------------------------------------------

max_payload_size = 512
max_request_size = 512

fmt_dict = {
    "mem_rd32": 0b00,
    "mem_rd64": 0b01,
    "mem_wr32": 0b10,
    "mem_wr64": 0b11,
    "cpld":     0b10,
    "cpl":      0b00,
}

type_dict = {
    "mem_rd32": 0b00000,
    "mem_rd64": 0b00000,
    "mem_wr32": 0b00000,
    "mem_wr64": 0b00000,
    "cpld":     0b01010,
    "cpl":      0b01010,
}

fmt_type_dict = {
    "mem_rd32": 0b00_00000,
    "mem_rd64": 0b01_00000,
    "mem_wr32": 0b10_00000,
    "mem_wr64": 0b11_00000,
    "cpld":     0b10_01010,
    "cpl":      0b00_01010,
}

cpl_dict = {
    "sc":  0b000,
    "ur":  0b001,
    "crs": 0b010,
    "ca":  0b011,
}

# Headers ------------------------------------------------------------------------------------------

tlp_common_header_length = 16
tlp_common_header_fields = {
    "fmt":  HeaderField(byte=0*4, offset=29, width=2),
    "type": HeaderField(byte=0*4, offset=24, width=5),
}
tlp_common_header = Header(
    fields           = tlp_common_header_fields,
    length           = tlp_common_header_length,
    swap_field_bytes = False
)

tlp_request_header_length = 16
tlp_request_header_fields = {
    "fmt":          HeaderField(byte=0*4, offset=29, width= 2),
    "type":         HeaderField(byte=0*4, offset=24, width= 5),
    "tc":           HeaderField(byte=0*4, offset=20, width= 3),
    "td":           HeaderField(byte=0*4, offset=15, width= 1),
    "ep":           HeaderField(byte=0*4, offset=14, width= 1),
    "attr":         HeaderField(byte=0*4, offset=12, width= 2),
    "length":       HeaderField(byte=0*4, offset= 0, width=10),

    "requester_id": HeaderField(byte=1*4, offset=16, width=16),
    "tag":          HeaderField(byte=1*4, offset= 8, width= 8),
    "last_be":      HeaderField(byte=1*4, offset= 4, width= 4),
    "first_be":     HeaderField(byte=1*4, offset= 0, width= 4),

    "address":      HeaderField(byte=2*4, offset= 0, width=64),
}
tlp_request_header = Header(
    fields           = tlp_request_header_fields,
    length           = tlp_request_header_length,
    swap_field_bytes = False
)

tlp_completion_header_length = 16
tlp_completion_header_fields = {
    "fmt":           HeaderField(byte=0*4, offset=29, width= 2),
    "type":          HeaderField(byte=0*4, offset=24, width= 5),
    "tc":            HeaderField(byte=0*4, offset=20, width= 3),
    "td":            HeaderField(byte=0*4, offset=15, width= 1),
    "ep":            HeaderField(byte=0*4, offset=14, width= 1),
    "attr":          HeaderField(byte=0*4, offset=12, width= 2),
    "length":        HeaderField(byte=0*4, offset= 0, width=10),

    "completer_id":  HeaderField(byte=1*4, offset=16, width=16),
    "status":        HeaderField(byte=1*4, offset=13, width= 3),
    "bcm":           HeaderField(byte=1*4, offset=12, width= 1),
    "byte_count":    HeaderField(byte=1*4, offset= 0, width=12),

    "requester_id":  HeaderField(byte=2*4, offset=16, width=16),
    "tag":           HeaderField(byte=2*4, offset= 8, width= 8),
    "lower_address": HeaderField(byte=2*4, offset= 0, width= 7),
}
tlp_completion_header = Header(
    fields           = tlp_completion_header_fields,
    length           = tlp_completion_header_length,
    swap_field_bytes = False
)

# Helpers ------------------------------------------------------------------------------------------

def dword_endianness_swap(src, dst, data_width, endianness, mode="dat", ndwords=None):
    assert len(src) == len(dst)
    assert data_width%32 == 0
    assert mode in ["dat", "be"]
    r = []
    nbits       = {"dat":            32, "be":            4}[mode]
    ndwords     = data_width//32 if ndwords is None else ndwords
    reverse_cls = {"dat": reverse_bytes, "be": reverse_bits}[mode]
    for n in range(ndwords):
        low  = (n + 0)*nbits
        high = (n + 1)*nbits
        r += {
            "little" : [dst[low:high].eq(            src[low:high])],
            "big"    : [dst[low:high].eq(reverse_cls(src[low:high]))],
        }[endianness]
    return r

# Layouts ------------------------------------------------------------------------------------------

def tlp_raw_layout(data_width):
    layout = [
        ("fmt",    2),
        ("header", 4*32),
        ("dat",    data_width),
        ("be",     data_width//8)
    ]
    return EndpointDescription(layout)


def tlp_common_layout(data_width):
    layout = tlp_common_header.get_layout() + [
        ("dat", data_width),
        ("be",  data_width//8)
    ]
    return EndpointDescription(layout)


def tlp_request_layout(data_width):
    layout = tlp_request_header.get_layout() + [
        ("dat", data_width),
        ("be",  data_width//8)
    ]
    return EndpointDescription(layout)


def tlp_completion_layout(data_width):
    layout = tlp_completion_header.get_layout() + [
        ("dat", data_width),
        ("be",  data_width//8)
    ]
    return EndpointDescription(layout)



#!/usr/bin/env python3

#
# This file is part of LitePCIe.
#
# Copyright (c) 2019-2023 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

"""
LitePCIe standalone core generator

LitePCIe aims to be directly used as a python package when the SoC is created using LiteX. However,
for some use cases it could be interesting to generate a standalone verilog file of the core:
- integration of the core in a SoC using a more traditional flow.
- need to version/package the core.
- avoid Migen/LiteX dependencies.
- etc...

The standalone core is generated from a YAML configuration file that allows the user to generate
easily a custom configuration of the core.

Current version of the generator is limited to:
- Xilinx 7-Series.
- Xilinx Ultrascale.
- Altera Cyclone V.
"""

import yaml
import argparse
import subprocess

from migen import *
from migen.genlib.resetsync import AsyncResetSynchronizer
from migen.genlib.misc import WaitTimer

from litex.gen import *

from litex.soc.cores.clock import *
from litex.soc.interconnect.csr import *
from litex.soc.interconnect import wishbone
from litex.soc.interconnect.axi import *
from litex.soc.integration.soc import SoCRegion
from litex.soc.integration.soc_core import *
from litex.soc.integration.builder import *

from litepcie.phy.c5pciephy import C5PCIEPHY
from litepcie.phy.s7pciephy import S7PCIEPHY
from litepcie.phy.uspciephy import USPCIEPHY
from litepcie.phy.usppciephy import USPPCIEPHY

from litepcie.core import LitePCIeEndpoint, LitePCIeMSI, LitePCIeMSIMultiVector, LitePCIeMSIX
from litepcie.frontend.dma import LitePCIeDMA
from litepcie.frontend.wishbone import LitePCIeWishboneMaster, LitePCIeWishboneSlave
from litepcie.frontend.axi import LitePCIeAXISlave
from litepcie.software import generate_litepcie_software_headers

from litex.build.generic_platform import *

# IOs/Interfaces -----------------------------------------------------------------------------------

def get_clk_ios():
    return [
        # Clk / Rst.
        ("clk", 0, Pins(1)),
        ("rst", 0, Pins(1))
    ]

def get_pcie_ios(phy_lanes=4):
    return [
        ("pcie", 0,
            Subsignal("rst_n", Pins(1)),
            Subsignal("clk_p", Pins(1)),
            Subsignal("clk_n", Pins(1)),
            Subsignal("rx_p",  Pins(phy_lanes)),
            Subsignal("rx_n",  Pins(phy_lanes)),
            Subsignal("tx_p",  Pins(phy_lanes)),
            Subsignal("tx_n",  Pins(phy_lanes)),
        ),
    ]

def get_axi_dma_ios(_id, data_width, with_writer=True, with_reader=True):
    ios = []

    # Enables.
    enables = []
    if with_writer:
        enables += [Subsignal("writer_enable", Pins(1))]
    if with_reader:
        enables += [Subsignal("reader_enable", Pins(1))]
    ios += [("dma{}_status".format(_id), 0, *enables)]

    # DMA Writer AXI.
    if with_writer:
        ios += [
            ("dma{}_writer_axi".format(_id), 0,
                Subsignal("tvalid", Pins(1)),
                Subsignal("tready", Pins(1)),
                Subsignal("tlast",  Pins(1)),
                Subsignal("tdata",  Pins(data_width)),
                Subsignal("tuser",  Pins(1)), # Use for tfirst.
            )
        ]

    # DMA Reader AXI.
    if with_reader:
        ios += [
            ("dma{}_reader_axi".format(_id), 0,
                Subsignal("tvalid", Pins(1)),
                Subsignal("tready", Pins(1)),
                Subsignal("tlast",  Pins(1)),
                Subsignal("tdata",  Pins(data_width)),
                Subsignal("tuser",  Pins(1)), # Use for tfirst.
            )
        ]

    return ios

def get_msi_irqs_ios(width=16):
    return [("msi_irqs", 0, Pins(width))]

# CRG ----------------------------------------------------------------------------------------------

class LitePCIeCRG(LiteXModule):
    def __init__(self, platform, sys_clk_freq, clk_external):
        self.cd_sys = ClockDomain()

        # # #

        # Create/Get Clk/Rst IOs.
        platform.add_extension(get_clk_ios())
        clk = platform.request("clk")
        rst = platform.request("rst")

        # Get PCIe Clk/Rst.
        pcie_clk = ClockSignal("pcie")
        pcie_rst = ResetSignal("pcie")

        # External Clk mode: Clk is provided by the User logic.
        if clk_external:
            self.comb += self.cd_sys.clk.eq(clk)
            self.specials += AsyncResetSynchronizer(self.cd_sys, rst | pcie_rst)

        # Internal Clk mode: Clk is provided to the User logic by the LitePCIe standalone core.
        else:
            self.comb += [
                clk.eq(pcie_clk),
                rst.eq(pcie_rst),
                self.cd_sys.clk.eq(pcie_clk),
                self.cd_sys.rst.eq(pcie_rst),
            ]

# Core ---------------------------------------------------------------------------------------------

class LitePCIeCore(SoCMini):
    SoCMini.mem_map["csr"] = 0x00000000
    SoCMini.csr_map = {
        "ctrl":           0,
        "crg" :           1,
        "pcie_phy":       2,
        "pcie_msi":       3,
        "pcie_msi_table": 4,
    }
    def __init__(self, platform, core_config):
        platform.add_extension(get_pcie_ios(core_config["phy_lanes"]))
        platform.add_extension(get_msi_irqs_ios(width=core_config["msi_irqs"]))
        sys_clk_freq = float(core_config.get("clk_freq", 125e6))

        # Parameters -------------------------------------------------------------------------------

        ep_address_width        = core_config.get("ep_address_width", 32)
        ep_max_pending_requests = core_config.get("ep_max_pending_requests", 4)

        # SoCMini ----------------------------------------------------------------------------------
        SoCMini.__init__(self, platform, clk_freq=sys_clk_freq,
            csr_data_width = 32,
            csr_ordering   = core_config.get("csr_ordering", "big"),
            ident          = "LitePCIe standalone core",
            ident_version  = True
        )

        # CRG --------------------------------------------------------------------------------------
        clk_external = core_config.get("clk_external", False)
        self.crg = LitePCIeCRG(platform, sys_clk_freq, clk_external)

        # Control ----------------------------------------------------------------------------------
        if core_config.get("ctrl", False):
            axi = AXILiteInterface(data_width=32, address_width=32)
            platform.add_extension(axi.get_ios("ctrl_axi_lite"))
            axi_pads = platform.request("ctrl_axi_lite")
            self.comb += axi.connect_to_pads(axi_pads, mode="slave")
            self.bus.add_master(name="ctrl", master=axi)

        # PCIe PHY ---------------------------------------------------------------------------------
        self.pcie_phy = core_config["phy"](platform, platform.request("pcie"),
            pcie_data_width = core_config.get("phy_pcie_data_width", 64),
            data_width      = core_config["phy_data_width"],
            bar0_size       = core_config["phy_bar0_size"])

        # PCIe Endpoint ----------------------------------------------------------------------------
        self.pcie_endpoint = LitePCIeEndpoint(self.pcie_phy,
            endianness           = self.pcie_phy.endianness,
            address_width        = ep_address_width,
            max_pending_requests = ep_max_pending_requests,
        )

        # PCIe Wishbone Master ---------------------------------------------------------------------
        pcie_wishbone_master = LitePCIeWishboneMaster(self.pcie_endpoint,
            qword_aligned = self.pcie_phy.qword_aligned)
        self.submodules += pcie_wishbone_master
        self.bus.add_master(master=pcie_wishbone_master.wishbone)

        # PCIe MMAP Master -------------------------------------------------------------------------
        if core_config.get("mmap", False):
            mmap_base        = core_config["mmap_base"]
            mmap_size        = core_config["mmap_size"]
            mmap_translation = core_config.get("mmap_translation", 0x00000000)
            mmap_region      = SoCRegion(origin=mmap_base, size=mmap_size, cached=False)
            wb = wishbone.Interface(data_width=32)
            self.bus.add_slave(name="mmap", slave=wb, region=mmap_region)
            axi = AXILiteInterface(data_width=32, address_width=32)
            wb2axi = Wishbone2AXILite(wb, axi, base_address=-mmap_translation)
            self.submodules += wb2axi
            platform.add_extension(axi.get_ios("mmap_axi_lite"))
            axi_pads = platform.request("mmap_axi_lite")
            self.comb += axi.connect_to_pads(axi_pads, mode="master")

        # PCIe MMAP Slave --------------------------------------------------------------------------
        if core_config.get("mmap_slave", False):
            # AXI-Full
            if core_config.get("mmap_slave_axi_full", False):
                pcie_axi_slave = LitePCIeAXISlave(self.pcie_endpoint, data_width=128)
                self.submodules += pcie_axi_slave
                platform.add_extension(pcie_axi_slave.axi.get_ios("mmap_slave_axi"))
                axi_pads = platform.request("mmap_slave_axi")
                self.comb += pcie_axi_slave.axi.connect_to_pads(axi_pads, mode="slave")
            # AXI-Lite
            else:
                platform.add_extension(axi.get_ios("mmap_slave_axi_lite"))
                axi_pads = platform.request("mmap_slave_axi_lite")
                wb = wishbone.Interface(data_width=32)
                axi = AXILiteInterface(data_width=32, address_width=32)
                self.comb += axi.connect_to_pads(axi_pads, mode="slave")
                axi2wb = AXILite2Wishbone(axi, wb)
                self.submodules += axi2wb
                pcie_wishbone_slave = LitePCIeWishboneSlave(self.pcie_endpoint,
                    qword_aligned=self.pcie_phy.qword_aligned)
                self.submodules += pcie_wishbone_slave
                self.comb += wb.connect(pcie_wishbone_slave.wishbone)

        # PCIe DMA ---------------------------------------------------------------------------------
        with_writer = core_config.get("dma_writer", True)
        with_reader = core_config.get("dma_reader", True)

        pcie_dmas = []
        self.add_constant("DMA_CHANNELS",   core_config["dma_channels"])
        self.add_constant("DMA_ADDR_WIDTH", ep_address_width)
        for i in range(core_config["dma_channels"]):
            pcie_dma = LitePCIeDMA(self.pcie_phy, self.pcie_endpoint,
                address_width     = ep_address_width,
                with_writer       = with_writer,
                with_reader       = with_reader,
                with_buffering    = core_config["dma_buffering"] != 0,
                buffering_depth   = core_config["dma_buffering"],
                with_loopback     = core_config["dma_loopback"],
                with_synchronizer = core_config["dma_synchronizer"],
                with_monitor      = core_config["dma_monitor"],
                with_writer_splitter_buffer = False, # FIXME: Prevent early termination, so don't use when early termination (tlast) is required.
                with_reader_splitter_buffer = True,
            )
            pcie_dma = stream.BufferizeEndpoints({"sink"   : stream.DIR_SINK})(pcie_dma)
            pcie_dma = stream.BufferizeEndpoints({"source" : stream.DIR_SOURCE})(pcie_dma)
            setattr(self.submodules, "pcie_dma" + str(i), pcie_dma)
            platform.add_extension(get_axi_dma_ios(i,
                data_width  = core_config["phy_data_width"],
                with_writer = with_writer,
                with_reader = with_reader,
            ))
            dma_status_ios = platform.request(f"dma{i}_status")

            if hasattr(pcie_dma, "writer"):
                dma_writer_ios = platform.request("dma{}_writer_axi".format(i))
                self.comb += [
                    # Status IOs
                    dma_status_ios.writer_enable.eq(pcie_dma.writer.enable),

                    # Writer IOs
                    pcie_dma.sink.valid.eq(dma_writer_ios.tvalid & pcie_dma.writer.enable),
                    dma_writer_ios.tready.eq(pcie_dma.sink.ready & pcie_dma.writer.enable),
                    pcie_dma.sink.last.eq(dma_writer_ios.tlast),
                    pcie_dma.sink.data.eq(dma_writer_ios.tdata),
                    pcie_dma.sink.first.eq(dma_writer_ios.tuser),
                ]
            if hasattr(pcie_dma, "reader"):
                dma_reader_ios = platform.request("dma{}_reader_axi".format(i))
                self.comb += [
                    # Status IOs
                    dma_status_ios.reader_enable.eq(pcie_dma.reader.enable),

                    # Reader IOs
                    dma_reader_ios.tvalid.eq(pcie_dma.source.valid & pcie_dma.reader.enable),
                    pcie_dma.source.ready.eq(dma_reader_ios.tready | ~pcie_dma.reader.enable),
                    dma_reader_ios.tlast.eq(pcie_dma.source.last),
                    dma_reader_ios.tdata.eq(pcie_dma.source.data),
                    dma_reader_ios.tuser.eq(pcie_dma.source.first),
                ]

        # PCIe MSI ---------------------------------------------------------------------------------
        if core_config.get("msi_x", False):
            assert core_config["msi_irqs"] <= 32
            self.pcie_msi = LitePCIeMSIX(self.pcie_endpoint, width=64)
            self.comb += self.pcie_msi.irqs[32:32+core_config["msi_irqs"]].eq(platform.request("msi_irqs"))
        else:
            assert core_config["msi_irqs"] <= 16
            if core_config.get("msi_multivector", False):
                self.pcie_msi = LitePCIeMSIMultiVector(width=32)
            else:
                self.pcie_msi = LitePCIeMSI(width=32)
            # On Ultrascale/Ultrascale+ limit rate of IRQs to 1MHz (to prevent issue with IRQs stalled).
            if isinstance(self.pcie_phy, (USPCIEPHY, USPPCIEPHY)):
                self.pcie_msi_timer = WaitTimer(int(sys_clk_freq/1e6))
                self.comb += self.pcie_msi_timer.wait.eq(~self.pcie_msi_timer.done)
                self.comb += If(self.pcie_msi_timer.done, self.pcie_msi.source.connect(self.pcie_phy.msi))
            else:
                self.comb += self.pcie_msi.source.connect(self.pcie_phy.msi)
            self.comb += self.pcie_msi.irqs[16:16+core_config["msi_irqs"]].eq(platform.request("msi_irqs"))
        self.interrupts = {}
        for i in range(core_config["dma_channels"]):
            pcie_dma = getattr(self, f"pcie_dma{i}")
            if hasattr(pcie_dma, "writer"):
                self.interrupts[f"pcie_dma{i}_writer"] = pcie_dma.writer.irq
            if hasattr(pcie_dma, "reader"):
                self.interrupts[f"pcie_dma{i}_reader"] = pcie_dma.reader.irq
        for i, (k, v) in enumerate(sorted(self.interrupts.items())):
            self.comb += self.pcie_msi.irqs[i].eq(v)
            self.add_constant(k.upper() + "_INTERRUPT", i)
        assert len(self.interrupts.keys()) <= 16

    def generate_documentation(self, build_name, **kwargs):
        from litex.soc.doc import generate_docs
        generate_docs(self, "documentation".format(build_name),
            project_name = "LitePCIe standalone core",
            author       = "Enjoy-Digital")
        os.system("sphinx-build -M html documentation/ documentation/_build".format(build_name, build_name))

# Build --------------------------------------------------------------------------------------------

def main():
    parser = argparse.ArgumentParser(description="LitePCIe standalone core generator")
    parser.add_argument("config", help="YAML config file")
    parser.add_argument("--doc",  action="store_true", help="Build documentation")
    args = parser.parse_args()
    core_config = yaml.load(open(args.config).read(), Loader=yaml.Loader)

    # Convert YAML elements to Python/LiteX --------------------------------------------------------
    for k, v in core_config.items():
        replaces = {"False": False, "True": True, "None": None}
        for r in replaces.keys():
            if v == r:
                core_config[k] = replaces[r]

    # Generate core --------------------------------------------------------------------------------
    if core_config["phy"]  == "C5PCIEPHY":
        from litex.build.altera import AlteraPlatform
        platform = AlteraPlatform("", io=[])
        core_config["phy"] = C5PCIEPHY
    elif core_config["phy"] == "S7PCIEPHY":
        from litex.build.xilinx import XilinxPlatform
        platform = XilinxPlatform(core_config["phy_device"], io=[], toolchain="vivado")
        core_config["phy"] = S7PCIEPHY
    elif core_config["phy"] == "USPCIEPHY":
        from litex.build.xilinx import XilinxPlatform
        platform = XilinxPlatform(core_config["phy_device"], io=[], toolchain="vivado")
        core_config["phy"] = USPCIEPHY
    elif core_config["phy"] == "USPPCIEPHY":
        from litex.build.xilinx import XilinxPlatform
        platform = XilinxPlatform(core_config["phy_device"], io=[], toolchain="vivado")
        core_config["phy"] = USPPCIEPHY
    else:
        raise ValueError("Unsupported PCIe PHY: {}".format(core_config["phy"]))
    soc      = LitePCIeCore(platform, core_config)
    builder  = Builder(soc, output_dir="build", compile_gateware=False)
    builder.build(build_name="litepcie_core", regular_comb=True)
    generate_litepcie_software_headers(soc, "./")

    if args.doc:
        soc.generate_documentation("litepcie_core")

if __name__ == "__main__":
    main()
